@@ -1,6 +1,23 @@
+/* This file is part of the KDE Project
+   Copyright (C) 1999 Klaas Freitag <freitag@suse.de>
 
-/** HEADER **/
+   This library is free software; you can redistribute it and/or
+   modify it under the terms of the GNU Library General Public
+   License as published by the Free Software Foundation; either
+   version 2 of the License, or (at your option) any later version.
 
+   This library is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+   Library General Public License for more details.
+
+   You should have received a copy of the GNU Library General Public License
+   along with this library; see the file COPYING.LIB.  If not, write to
+   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
+   Boston, MA 02111-1307, USA.
+
+   $Id: kscandevice.cpp,v 1.43 2003/08/25 14:05:01 freitag Exp $
+*/
 
 #include <stdlib.h>
 #include <qwidget.h>
@@ -28,21 +45,11 @@
 #include "kscanoption.h"
 #include "kscanoptset.h"
 #include "devselector.h"
+#include <ksimpleconfig.h>
 
 #define MIN_PREVIEW_DPI 75
-
-
-// global device list
-
-/** These variables are defined in ksaneoption.cpp **/
-extern bool        scanner_initialised;
-extern SANE_Handle scanner_handle;
-extern QAsciiDict<int>  option_dic;
-
-extern SANE_Device const **dev_list          = 0;
-
-KScanOptSet gammaTables("GammaTables");
-
+#define UNDEF_SCANNERNAME I18N_NOOP( "undefined" )
+#define MAX_PROGRESS 100
 
 /* ---------------------------------------------------------------------------
 
@@ -69,7 +76,12 @@ KScanOption *KScanDevice::getExistingGuiElement( const QCString& name )
     KScanOption *ret = 0L;
 
     QCString alias = aliasName( name );
-    ret = gui_elem_names[ alias ];
+
+    /* gui_elements is a QList<KScanOption> */
+    for( ret = gui_elements.first(); ret != 0; ret = gui_elements.next())
+    {
+       if( ret->getName() == alias ) break;
+    }
 
     return( ret );
 }
@@ -99,12 +111,18 @@ KScanOption *KScanDevice::getGuiElement( const QCString& name, QWidget *parent,
    {
       /** store new gui-elem in list of all gui-elements */
       gui_elements.append( so );
-      gui_elem_names.insert( alias, so );
-	 	
+
       w = so->createWidget( parent, desc, tooltip );
       if( w )
+      {
 	 connect( so,   SIGNAL( optionChanged( KScanOption* ) ),
-		  this, SLOT(   slOptChanged( KScanOption* )));	
+		  this, SLOT(   slOptChanged( KScanOption* )));
+	 w->setEnabled( so->active() );
+      }
+      else
+      {
+	 kdDebug(29000) << "ERROR: No widget created for " << name << endl;
+      }
    }
    else
    {
@@ -117,7 +135,7 @@ KScanOption *KScanDevice::getGuiElement( const QCString& name, QWidget *parent,
       delete so;
       so = 0;
    }
-	
+
    return( so );
 }
 // ---------------------------------------------------------------------------
@@ -130,7 +148,8 @@ KScanDevice::KScanDevice( QObject *parent )
 {
     SANE_Status sane_stat = sane_init(NULL, NULL );
 
-    option_dic.setAutoDelete( true );
+    option_dic = new QAsciiDict<int>;
+    option_dic->setAutoDelete( true );
     gui_elements.setAutoDelete( true );
 
     scanner_initialised = false;  /* stays false until openDevice. */
@@ -170,6 +189,7 @@ KScanDevice::KScanDevice( QObject *parent )
 #if 0
         connect( this, SIGNAL(sigOptionsChanged()), SLOT(slReloadAll()));
 #endif
+	gammaTables = new KScanOptSet( "GammaTables" );
      }
      else
      {
@@ -184,6 +204,8 @@ KScanDevice::KScanDevice( QObject *parent )
 KScanDevice::~KScanDevice()
 {
   if( storeOptions )  delete (storeOptions );
+  kdDebug(29000) << "Calling sane_exit to finish sane!" << endl;
+  sane_exit();
 }
 
 
@@ -212,10 +234,10 @@ KScanStat KScanDevice::openDevice( const QCString& backend )
 	     // fill description dic with names and numbers
 	stat = find_options();
 	scanner_name = backend;
-	
+
       } else {
 	stat = KSCAN_ERR_OPEN_DEV;
-	scanner_name = "undefined";
+	scanner_name = UNDEF_SCANNERNAME;
       }
    }
 
@@ -225,6 +247,34 @@ KScanStat KScanDevice::openDevice( const QCString& backend )
    return( stat );
 }
 
+void KScanDevice::slCloseDevice( )
+{
+   /* First of all, send a signal to close down the scanner dev. */
+   emit( sigCloseDevice( ));
+
+   kdDebug(29000) << "Saving scan settings" << endl;
+   slSaveScanConfigSet( DEFAULT_OPTIONSET, i18n("the default startup setup"));
+
+   /* After return, delete all related stuff */
+   scanner_name = UNDEF_SCANNERNAME;
+   if( scanner_handle )
+   {
+      if( scanStatus != SSTAT_SILENT )
+      {
+         kdDebug(29000) << "Scanner is still active, calling cancel !" << endl;
+         sane_cancel( scanner_handle );
+      }
+      sane_close( scanner_handle );
+      scanner_handle = 0;
+   }
+
+   gui_elements.clear();
+
+   option_dic->clear();
+   scanner_initialised = false;
+
+}
+
 
 QString KScanDevice::getScannerName(const QCString& name) const
 {
@@ -242,8 +292,9 @@ QString KScanDevice::getScannerName(const QCString& name) const
   }
 
   if( scanner ) {
-     ret.sprintf( "%s %s %s", scanner->vendor, scanner->model, scanner->type );
-  }		
+     // ret.sprintf( "%s %s %s", scanner->vendor, scanner->model, scanner->type );
+     ret.sprintf( "%s %s", scanner->vendor, scanner->model );
+  }
 
   kdDebug(29000) << "getScannerName returns <" << ret << ">" << endl;
   return ( ret );
@@ -288,7 +339,7 @@ KScanStat KScanDevice::find_options()
   if( stat == KSCAN_OK )
   {
 
-     option_dic.clear();
+     option_dic->clear();
 
      for(int i = 1; i<n; i++)
      {
@@ -308,14 +359,14 @@ KScanStat KScanDevice::find_options()
 		 *new_opt = i;
 		 kdDebug(29000) << "Inserting <" << d->name << "> as " << *new_opt << endl;
 		 /* create a new option in the set. */
-		 option_dic.insert ( (const char*)d->name, new_opt );
+		 option_dic->insert ( (const char*)d->name, new_opt );
 		 option_list.append( (const char*) d->name );
 #if 0
 		 KScanOption *newOpt = new KScanOption( d->name );
 		 const QString qq = newOpt->get();
 		 qDebug( "INIT: <%s> = <%s>", d->name, qq );
 		 allOptionSet->insert( d->name, newOpt );
-#endif		
+#endif
 	      }
 	      else if( d->type == SANE_TYPE_GROUP )
 	      {
@@ -374,7 +425,7 @@ KScanStat KScanDevice::apply( KScanOption *opt, bool isGammaTable )
    if( !opt ) return( KSCAN_ERR_PARAM );
    int sane_result = 0;
 
-   int         *num = option_dic[ opt->getName() ];
+   int         *num = (*option_dic)[ opt->getName() ];
    SANE_Status sane_stat = SANE_STATUS_GOOD;
    const QCString& oname = opt->getName();
 
@@ -389,7 +440,7 @@ KScanStat KScanDevice::apply( KScanOption *opt, bool isGammaTable )
    if( ! opt->initialised() || opt->getBuffer() == 0 )
    {
       kdDebug(29000) << "Attempt to set Zero buffer of " << oname << " -> skipping !" << endl;
-   	
+
       if( opt->autoSetable() )
       {
 	 kdDebug(29000) << "Setting option automatic !" << endl;
@@ -404,7 +455,7 @@ KScanStat KScanDevice::apply( KScanOption *opt, bool isGammaTable )
       stat = KSCAN_ERR_PARAM;
    }
    else
-   {	
+   {
       if( ! opt->active() )
       {
 	 kdDebug(29000) << "Option " << oname << " is not active now!" << endl;
@@ -417,13 +468,13 @@ KScanStat KScanDevice::apply( KScanOption *opt, bool isGammaTable )
       }
       else
       {
-	
+
 	 sane_stat = sane_control_option( scanner_handle, *num,
 					  SANE_ACTION_SET_VALUE,
 					  opt->getBuffer(),
 					  &sane_result );
       }
-   }		
+   }
 
    if( stat == KSCAN_OK )
    {
@@ -453,7 +504,7 @@ KScanStat KScanDevice::apply( KScanOption *opt, bool isGammaTable )
 	 /* if it is a gamma table, the gamma values must be stored */
 	 if( isGammaTable )
 	 {
-	    gammaTables.backupOption( *opt );
+	    gammaTables->backupOption( *opt );
 	    kdDebug(29000) << "GammaTable stored: " << opt->getName() << endl;
 	 }
       }
@@ -468,6 +519,12 @@ KScanStat KScanDevice::apply( KScanOption *opt, bool isGammaTable )
    {
       kdDebug(29000) << "Setting of <" << oname << "> failed -> kscanerror." << endl;
    }
+
+   if( stat == KSCAN_OK )
+   {
+      slSetDirty( oname );
+   }
+
    return( stat );
 }
 
@@ -479,13 +536,27 @@ bool KScanDevice::optionExists( const QCString& name )
    QCString altname = aliasName( name );
 
    if( ! altname.isNull() )
-       i = option_dic[ altname ];
+       i = (*option_dic)[ altname ];
 
    if( !i )
        return( false );
    return( *i > -1 );
 }
 
+void KScanDevice::slSetDirty( const QCString& name )
+{
+   if( optionExists( name ) )
+   {
+      if( dirtyList.find( name ) == -1 )
+      {
+	 kdDebug(29000)<< "Setting dirty <" << name << ">" << endl;
+	 /* item not found */
+	 dirtyList.append( name );
+      }
+   }
+}
+
+
 /* This function tries to find name aliases which appear from backend to backend.
  *  Example: Custom-Gamma is for epson backends 'gamma-correction' - not a really
  *  cool thing :-|
@@ -493,7 +564,7 @@ bool KScanDevice::optionExists( const QCString& name )
  */
 QCString KScanDevice::aliasName( const QCString& name )
 {
-    int *i = option_dic[ name ];
+    int *i = (*option_dic)[ name ];
     QCString ret;
 
     if( i ) return( name );
@@ -501,9 +572,9 @@ QCString KScanDevice::aliasName( const QCString& name )
 
     if( name == SANE_NAME_CUSTOM_GAMMA )
     {
-	if(option_dic["gamma-correction"])
+	if((*option_dic)["gamma-correction"])
 	    ret = "gamma-correction";
-	
+
     }
 
     if( ret != name )
@@ -528,13 +599,13 @@ void KScanDevice::slReloadAllBut( KScanOption *not_opt )
         {
             kdDebug(29000) << "ReloadAllBut called with invalid argument" << endl;
             return;
-        }        	
+        }
 
         /* Make sure its applied */
 	apply( not_opt );
 
 	kdDebug(29000) << "*** Reload of all except <" << not_opt->getName() << "> forced ! ***" << endl;
-	
+
 	for( KScanOption *so = gui_elements.first(); so; so = gui_elements.next())
 	{
 	    if( so != not_opt )
@@ -545,7 +616,7 @@ void KScanDevice::slReloadAllBut( KScanOption *not_opt )
 	    }
 	}
 	kdDebug(29000) << "*** Reload of all finished ! ***" << endl;
-		
+
 }
 
 
@@ -553,7 +624,7 @@ void KScanDevice::slReloadAllBut( KScanOption *not_opt )
 void KScanDevice::slReloadAll( )
 {
 	kdDebug(29000) << "*** Reload of all forced ! ***" << endl;
-	
+
 	for( KScanOption *so = gui_elements.first(); so; so = gui_elements.next())
 	{
 	 	so->slReload();
@@ -567,9 +638,9 @@ void KScanDevice::slStopScanning( void )
     kdDebug(29000) << "Attempt to stop scanning" << endl;
     if( scanStatus == SSTAT_IN_PROGRESS )
     {
-    	scanStatus = SSTAT_STOP_NOW;
 	emit( sigScanFinished( KSCAN_CANCELLED ));
     }
+    scanStatus = SSTAT_STOP_NOW;
 }
 
 KScanStat KScanDevice::acquirePreview( bool forceGray, int dpi )
@@ -578,21 +649,6 @@ KScanStat KScanDevice::acquirePreview( bool forceGray, int dpi )
 
    (void) forceGray;
 
-
-#if 0
-   QAsciiDictIterator<KScanOption> it(*standardOptions);
-
-   while( it.current() )
-   {
-      so = it.current();
-      if( so && so->active() && so->initialised())
-      {
-	 qDebug( "apply <%s>", (const char*) it.currentKey() );
-	 apply( so );
-      }
-      ++it;
-   }
-#endif
    if( storeOptions )
       storeOptions->clear();
    else
@@ -660,8 +716,8 @@ KScanStat KScanDevice::acquirePreview( bool forceGray, int dpi )
       if( ! res.getRange( &min, &max, &q ) )
       {
 	 kdDebug(29000) << "Could not retrieve resolution range!" << endl;
-	 min = 75.0; // Hope that every scanner can 75 
-      }      
+	 min = 75.0; // Hope that every scanner can 75
+      }
       if( min > MIN_PREVIEW_DPI )
 	 set_dpi = (int) min;
       else
@@ -708,18 +764,30 @@ KScanStat KScanDevice::acquirePreview( bool forceGray, int dpi )
  * prepareScan tries to set as much as parameters as possible.
  *
  **/
-#define NOTIFIER(X) ( X == 0  ? ' ' : 'X' )
+#define NOTIFIER(X) optionNotifyString(X)
+
+QString KScanDevice::optionNotifyString( int i ) const
+{
+    const QString sOff = "        |";
+    const QString sOn  = "   X    |";   
+    if( i > 0 )
+    {
+	return sOn;
+    }
+    return sOff;
+}
+
 
 void KScanDevice::prepareScan( void )
 {
-    QAsciiDictIterator<int> it( option_dic ); // iterator for dict
+    QAsciiDictIterator<int> it( *option_dic ); // iterator for dict
 
     kdDebug(29000) << "########################################################################################################" << endl;
     kdDebug(29000) << "Scanner: " << scanner_name << endl;
     kdDebug(29000) << "         " << getScannerName() << endl;
-    kdDebug(29000) << "----------------------+-----------+-----------+-----------+-----------+-----------+-----------+-----------+" << endl;
-    kdDebug(29000) << " Option-Name          | SOFT_SEL  | HARD_SEL  | SOFT_DET  | EMULATED  | AUTOMATIC | INACTIVE  | ADVANCED  |" << endl;
-    kdDebug(29000) << "----------------------+-----------+-----------+-----------+-----------+-----------+-----------+-----------+" << endl;
+    kdDebug(29000) << "----------------------------------+--------+--------+--------+--------+--------+--------+--------+" << endl;
+    kdDebug(29000) << " Option-Name                      |SOFT_SEL|HARD_SEL|SOFT_DET|EMULATED|AUTOMATI|INACTIVE|ADVANCED|" << endl;
+    kdDebug(29000) << "----------------------------------+--------+--------+--------+--------+--------+--------+--------+" << endl;
 
     while ( it.current() )
     {
@@ -731,20 +799,21 @@ void KScanDevice::prepareScan( void )
        if( d )
        {
 	  int cap = d->cap;
-
-	  kdDebug(29000) << " " << it.currentKey() << " |     " <<
-		 NOTIFIER( ((cap) & SANE_CAP_SOFT_SELECT) ) << " |     " <<
-		 NOTIFIER( ((cap) & SANE_CAP_HARD_SELECT) ) << " |     " <<
-		 NOTIFIER( ((cap) & SANE_CAP_SOFT_DETECT) ) << " |     " <<
-		 NOTIFIER( ((cap) & SANE_CAP_EMULATED) ) << " |     " <<
-		 NOTIFIER( ((cap) & SANE_CAP_AUTOMATIC) ) << " |     " <<
-		 NOTIFIER( ((cap) & SANE_CAP_INACTIVE) ) << " |     " <<
-		 NOTIFIER( ((cap) & SANE_CAP_ADVANCED ) ) << " |" << endl;
+	  
+	  QString s = QString(it.currentKey()).leftJustify(32, ' ');
+	  kdDebug(29000) << " " << s << " |" <<
+		 NOTIFIER( ((cap) & SANE_CAP_SOFT_SELECT)) << 
+		 NOTIFIER( ((cap) & SANE_CAP_HARD_SELECT)) << 
+		 NOTIFIER( ((cap) & SANE_CAP_SOFT_DETECT)) << 
+		 NOTIFIER( ((cap) & SANE_CAP_EMULATED)   ) << 
+		 NOTIFIER( ((cap) & SANE_CAP_AUTOMATIC)  ) << 
+		 NOTIFIER( ((cap) & SANE_CAP_INACTIVE)   ) << 
+		 NOTIFIER( ((cap) & SANE_CAP_ADVANCED )  )    << endl;
 
        }
        ++it;
     }
-    kdDebug(29000) << "----------------------+-----------+-----------+-----------+-----------+-----------+-----------+-----------+" << endl;
+    kdDebug(29000) << "----------------------------------+--------+--------+--------+--------+--------+--------+--------+" << endl;
 
     KScanOption pso( SANE_NAME_PREVIEW );
     const QString q = pso.get();
@@ -761,12 +830,12 @@ void KScanDevice::prepareScan( void )
 KScanStat KScanDevice::acquire( const QString& filename )
 {
     KScanOption *so = 0;
- 	
+
     if( filename.isEmpty() )
     {
  	/* *real* scanning - apply all Options and go for it */
  	prepareScan();
- 	
+
  	for( so = gui_elements.first(); so; so = gui_elements.next() )
  	{
  	    if( so->active() )
@@ -779,7 +848,7 @@ KScanStat KScanDevice::acquire( const QString& filename )
  	        kdDebug(29000) << "Option <" << so->getName() << "> is not active !" << endl;
  	    }
  	}
-	return( acquire_data( false ));	
+	return( acquire_data( false ));
     }
     else
     {
@@ -807,7 +876,10 @@ KScanStat KScanDevice::createNewImage( SANE_Parameters *p )
   if( !p ) return( KSCAN_ERR_PARAM );
   KScanStat       stat = KSCAN_OK;
 
-  if( img ) delete( img );
+  if( img ) {
+     delete( img );
+     img = 0;
+  }
 
   if( p->depth == 1 )  //  Lineart !!
   {
@@ -818,7 +890,7 @@ KScanStat KScanDevice::createNewImage( SANE_Parameters *p )
       img->setColor( 0, qRgb( 0,0,0));
       img->setColor( 1, qRgb( 255,255,255) );
     }
-  }	
+  }
   else if( p->depth == 8 )
   {
     // 8 bit rgb-Picture
@@ -829,7 +901,7 @@ KScanStat KScanDevice::createNewImage( SANE_Parameters *p )
       if( img )
       {
 	img->setNumColors( 256 );
-	
+
 	for(int i = 0; i<256; i++)
 	  img->setColor(i, qRgb(i,i,i));
       }
@@ -860,6 +932,8 @@ KScanStat KScanDevice::acquire_data( bool isPreview )
 
    scanningPreview = isPreview;
 
+   emit sigScanStart();
+
    sane_stat = sane_start( scanner_handle );
    if( sane_stat == SANE_STATUS_GOOD )
    {
@@ -867,6 +941,7 @@ KScanStat KScanDevice::acquire_data( bool isPreview )
 
       if( sane_stat == SANE_STATUS_GOOD )
       {
+	 kdDebug(29000) << "--Pre-Loop" << endl;
 	 kdDebug(29000) << "format : " << sane_scan_param.format << endl;
 	 kdDebug(29000) << "last_frame : " << sane_scan_param.last_frame << endl;
 	 kdDebug(29000) << "lines : " << sane_scan_param.lines << endl;
@@ -886,8 +961,6 @@ KScanStat KScanDevice::acquire_data( bool isPreview )
 	 kdDebug(29000) << "sane-start-Error: " << sane_strstatus( sane_stat ) << endl;
    }
 
-   /* Signal for a progress dialog */
-   emit( sigScanProgress( 0 ) );
 
    if( sane_scan_param.pixels_per_line == 0 || sane_scan_param.lines < 1 )
    {
@@ -909,11 +982,15 @@ KScanStat KScanDevice::acquire_data( bool isPreview )
       if( ! data ) stat = KSCAN_ERR_MEMORY;
    }
 
+   /* Signal for a progress dialog */
+   emit( sigScanProgress( 0 ) );
+   emit( sigAcquireStart() );
+
    if( stat == KSCAN_OK )
-   {	
+   {
       /* initiates Redraw of the Progress-Window */
       qApp->processEvents();
-   }	
+   }
 
    if( stat == KSCAN_OK )
    {
@@ -928,14 +1005,14 @@ KScanStat KScanDevice::acquire_data( bool isPreview )
        *  this status might be changed by pressing Stop on a GUI-Dialog displayed during scan */
       if( sane_set_io_mode( scanner_handle, SANE_TRUE ) == SANE_STATUS_GOOD )
       {
-	
+
 	 int fd = 0;
 	 if( sane_get_select_fd( scanner_handle, &fd ) == SANE_STATUS_GOOD )
 	 {
 	    sn = new QSocketNotifier( fd, QSocketNotifier::Read, this );
 	    QObject::connect( sn, SIGNAL(activated(int)),
 			      this, SLOT( doProcessABlock() ) );
-	
+
 	 }
       }
       else
@@ -946,7 +1023,7 @@ KScanStat KScanDevice::acquire_data( bool isPreview )
 	    if( scanStatus != SSTAT_SILENT )
 	    {
 	       sane_stat = sane_get_parameters( scanner_handle, &sane_scan_param );
-
+	       kdDebug(29000) << "--ProcessABlock-Loop" << endl;
 	       kdDebug(29000) << "format : " << sane_scan_param.format << endl;
 	       kdDebug(29000) << "last_frame : " << sane_scan_param.last_frame << endl;
 	       kdDebug(29000) << "lines : " << sane_scan_param.lines << endl;
@@ -967,6 +1044,37 @@ KScanStat KScanDevice::acquire_data( bool isPreview )
    return( stat );
 }
 
+void KScanDevice::loadOptionSet( KScanOptSet *optSet )
+{
+   if( !optSet )
+   {
+      return;
+   }
+
+   // kdDebug(29000) << "Loading Option set: " << optSet->optSetName() << endl;
+   QAsciiDictIterator<KScanOption> it(*optSet);
+
+   kdDebug(29000) << "Postinstalling " << optSet->count() << " options" << endl;
+   while( it.current() )
+   {
+      KScanOption *so = it.current();
+      if( ! so->initialised() )
+	 kdDebug(29000) << so->getName() << " is not initialised" << endl;
+
+      if( ! so->active() )
+	 kdDebug(29000) << so->getName() << " is not active" << endl;
+
+      if( so && so->active() && so->initialised())
+      {
+	 const QString qq = so->get();
+
+	 kdDebug(29000) << "Post-Scan-apply <" << it.currentKey() << ">: " << qq << endl;
+	 apply( so );
+      }
+      ++it;
+   }
+
+}
 
 
 void KScanDevice::slScanFinished( KScanStat status )
@@ -978,7 +1086,7 @@ void KScanDevice::slScanFinished( KScanStat status )
       sn = 0;
    }
 
-   emit( sigScanProgress( 1000 ));
+   emit( sigScanProgress( MAX_PROGRESS ));
 
    kdDebug(29000) << "Slot ScanFinished hit with status " <<  status << endl;
 
@@ -1000,30 +1108,13 @@ void KScanDevice::slScanFinished( KScanStat status )
 	 emit( sigNewPreview( img ));
 
 	 /* The old settings need to be redefined */
-	 QAsciiDictIterator<KScanOption> it(*storeOptions);
-
-	 kdDebug(29000) << "Postinstalling " << storeOptions->count() << " options" << endl;
-	 while( it.current() )
-	 {
-	    KScanOption *so = it.current();
-	    if( ! so->initialised() ) kdDebug(29000) << so->getName() << " is not initialised" << endl;
-	    if( ! so->active() ) kdDebug(29000) << so->getName() << " is not active" << endl;
-
-	    if( so && so->active() && so->initialised())
-	    {
-	       const QString qq = so->get();	
-	
-	       kdDebug(29000) << "Post-Scan-apply <" << it.currentKey() << ">: " << qq << endl;
-	       apply( so );
-	    }
-	    ++it;
-	 }
+	 loadOptionSet( storeOptions );
       }
       else
       {
 	 emit( sigNewImage( img ));
       }
-   }	
+   }
 
 
    sane_cancel(scanner_handle);
@@ -1099,7 +1190,7 @@ void KScanDevice::doProcessABlock( void )
 	      if( sane_scan_param.lines < 1 ) break;
 	      bytes_written += rest_bytes; // die bergebliebenen Bytes dazu
 	      rest_bytes = bytes_written % 3;
-	
+
 	      for( val = 0; val < ((bytes_written-rest_bytes) / 3); val++ )
 	      {
 		 red   = *rptr++;
@@ -1119,7 +1210,7 @@ void KScanDevice::doProcessABlock( void )
 	      {
 		 *(data+val) = *rptr++;
 	      }
-	
+
 	      break;
 	   case SANE_FRAME_GRAY:
 	      for( val = 0; val < bytes_written ; val++ )
@@ -1137,7 +1228,6 @@ void KScanDevice::doProcessABlock( void )
 		    eight_pix = *rptr++;
 		    for( i = 0; i < 8; i++ )
 		    {
-		       if( pixel_x  >= sane_scan_param.pixels_per_line ) { pixel_x = 0; pixel_y++; }
 		       if( pixel_y < sane_scan_param.lines )
 		       {
 			  chan = (eight_pix & 0x80)> 0 ? 0:1;
@@ -1145,11 +1235,16 @@ void KScanDevice::doProcessABlock( void )
 			    	//qDebug( "Setting on %d, %d: %d", pixel_x, pixel_y, chan );
 			  img->setPixel( pixel_x, pixel_y, chan );
 			  pixel_x++;
+			  if( pixel_x  >= sane_scan_param.pixels_per_line )
+			  {
+			     pixel_x = 0; pixel_y++;
+			     break;
+			  }
 		       }
 		    }
 		 }
-	      }	
-	
+	      }
+
 	      break;
 	   case SANE_FRAME_RED:
 	   case SANE_FRAME_GREEN:
@@ -1161,17 +1256,17 @@ void KScanDevice::doProcessABlock( void )
 		 {
 		    pixel_y++; pixel_x = 0;
 		 }
-	
+
 		 if( pixel_y < sane_scan_param.lines )
 		 {
-		
+
 		    col = img->pixel( pixel_x, pixel_y );
-		
+
 		    red   = qRed( col );
 		    green = qGreen( col );
 		    blue  = qBlue( col );
 		    chan  = *rptr++;
-		
+
 		    switch( sane_scan_param.format )
 		    {
 		       case SANE_FRAME_RED :
@@ -1198,12 +1293,12 @@ void KScanDevice::doProcessABlock( void )
 	      break;
 	} /* switch */
 
-	
+
 	if( (sane_scan_param.lines > 0) && (sane_scan_param.lines * pixel_y> 0) )
 	{
-	   int progress =  (int)(1000.0 / (double) sane_scan_param.lines *
+	   int progress =  (int)((double)MAX_PROGRESS / (double) sane_scan_param.lines *
 				 (double)pixel_y);
-	   if( progress < 1000 )
+	   if( progress < MAX_PROGRESS)
 	      emit( sigScanProgress( progress));
 	}
 
@@ -1223,8 +1318,10 @@ void KScanDevice::doProcessABlock( void )
 	 * the QSocketnotifier fires for a few times after the scan has been
 	 * cancelled.  Does it matter ? To see it, just uncomment the qDebug msg.
 	 */
-	// qDebug( "Stopping the scan progress !" );
+	kdDebug(29000) << "Stopping the scan progress !" << endl;
 	goOn = false;
+	scanStatus = SSTAT_SILENT;
+	emit( sigScanFinished( KSCAN_OK ));
      }
 
   } /* while( 1 ) */
@@ -1236,10 +1333,10 @@ void KScanDevice::doProcessABlock( void )
      if ( sane_scan_param.last_frame)
      {
 	/** Everythings okay, the picture is ready **/
-	kdDebug(29000) << "last frame reached - scan successfull" << endl;
+	kdDebug(29000) << "last frame reached - scan successful" << endl;
 	scanStatus = SSTAT_SILENT;
 	emit( sigScanFinished( KSCAN_OK ));
-     }	
+     }
      else
      {
 	/** EOF und nicht letzter Frame -> Parameter neu belegen und neu starten **/
@@ -1250,7 +1347,7 @@ void KScanDevice::doProcessABlock( void )
   }
 
   if( sane_stat == SANE_STATUS_CANCELLED )
-  {	
+  {
      scanStatus = SSTAT_STOP_NOW;
      kdDebug(29000) << "Scan was cancelled" << endl;
 
@@ -1266,16 +1363,78 @@ void KScanDevice::slSaveScanConfigSet( const QString& setName, const QString& de
 {
    if( setName.isEmpty() || setName.isNull()) return;
 
-   KScanOptSet optSet( "saveSet" );
+   kdDebug(29000) << "Saving Scan Configuration" << setName << endl;
 
-   for( KScanOption *so = gui_elements.first(); so; so = gui_elements.next())
-   {
-      kdDebug(29000) << "Storing <" << so->getName() << ">" << endl;
-      optSet.backupOption( *so );
-   }
-   optSet.saveConfig( "default" , setName, descr );
+   KScanOptSet optSet( DEFAULT_OPTIONSET );
+   getCurrentOptions( &optSet );
+
+   optSet.saveConfig( scanner_name , setName, descr );
 
 }
 
 
+void  KScanDevice::getCurrentOptions( KScanOptSet *optSet )
+{
+   if( ! optSet ) return;
+
+   for( KScanOption *so = gui_elements.first(); so; so = gui_elements.next())
+   {
+      kdDebug(29000) << "Storing <" << so->getName() << ">" << endl;
+      if( so && so->active())
+      {
+	 apply(so);
+	 optSet->backupOption( *so );
+      }
+
+      /* drop the thing from the dirty-list */
+      dirtyList.removeRef( so->getName());
+   }
+
+   QStrListIterator it( dirtyList );
+   while( it.current())
+   {
+      KScanOption so( it.current() );
+      optSet->backupOption( so );
+      ++it;
+   }
+}
+
+QString KScanDevice::getConfig( const QString& key, const QString& def ) const
+{
+    QString confFile = SCANNER_DB_FILE;
+
+    KSimpleConfig scanConfig( confFile, true );
+    scanConfig.setGroup( shortScannerName() );
+
+    return scanConfig.readEntry( key, def );
+}
+
+void KScanDevice::slStoreConfig( const QString& key, const QString& val )
+{
+    QString confFile = SCANNER_DB_FILE;
+    QString scannerName = shortScannerName();
+
+    if( scannerName.isEmpty() || scannerName == UNDEF_SCANNERNAME )
+    {
+        kdDebug(29000) << "Skipping config write, scanner name is empty!" << endl;
+    }
+    else
+    {
+        kdDebug(29000) << "Storing config " << key << " in Group " << scannerName << endl;
+
+        KSimpleConfig scanConfig( confFile );
+        scanConfig.setGroup( scannerName );
+        scanConfig.writeEntry( key, val );
+        scanConfig.sync();
+    }
+}
+
+
+bool KScanDevice::scanner_initialised = false;
+SANE_Handle KScanDevice::scanner_handle = 0L;
+SANE_Device const **KScanDevice::dev_list = 0L;
+QAsciiDict<int> *KScanDevice::option_dic = 0L;
+KScanOptSet *KScanDevice::gammaTables = 0L;
+
+
 #include "kscandevice.moc"
